@@ -27,6 +27,7 @@
 #include <gtk/gtk.h>
 #include <gmodule.h>
 #include <math.h>
+#include <cairo-gobject.h>
 
 #define ADWAITA_NAMESPACE "adwaita"
 
@@ -89,6 +90,32 @@ _cairo_round_rectangle (cairo_t *cr,
 	cairo_arc (cr, x+radius,   y+radius,   radius, G_PI, G_PI * 1.5);
 }
 
+/* Set the appropriate matrix for
+ * patterns coming from the style context
+ */
+static void
+style_pattern_set_matrix (cairo_pattern_t *pattern,
+                          gdouble          width,
+                          gdouble          height)
+{
+  cairo_matrix_t matrix;
+  gint w, h;
+
+  if (cairo_pattern_get_type (pattern) == CAIRO_PATTERN_TYPE_SURFACE)
+    {
+      cairo_surface_t *surface;
+
+      cairo_pattern_get_surface (pattern, &surface);
+      w = cairo_image_surface_get_width (surface);
+      h = cairo_image_surface_get_height (surface);
+    }
+  else
+    w = h = 1;
+
+  cairo_matrix_init_scale (&matrix, (gdouble) w / width, (gdouble) h / height);
+  cairo_pattern_set_matrix (pattern, &matrix);
+}
+
 static void
 adwaita_engine_render_arrow (GtkThemingEngine *engine,
 			     cairo_t	      *cr,
@@ -122,6 +149,22 @@ adwaita_engine_render_arrow (GtkThemingEngine *engine,
 
         state = gtk_theming_engine_get_state (engine);
 	gtk_theming_engine_get_color (engine, state, &color);
+
+	if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_MENUITEM) &&
+	    !(state & GTK_STATE_FLAG_INSENSITIVE) && !(state & GTK_STATE_FLAG_PRELIGHT)) {
+		GdkRGBA *arrow_color;
+
+		gtk_theming_engine_get (engine, state,
+					"-adwaita-menuitem-arrow-color", &arrow_color,
+					NULL);
+
+		if (arrow_color != NULL) {
+			color = *arrow_color;
+		}
+
+		gdk_rgba_free (arrow_color);
+	}
+
 	gdk_cairo_set_source_rgba (cr, &color);
 	cairo_stroke (cr);
         
@@ -137,25 +180,36 @@ adwaita_engine_render_focus (GtkThemingEngine *engine,
 			     gdouble	       height)
 {
 	GdkRGBA *fill_color, *border_color = NULL;
-	GdkRGBA *border_gradient_a = NULL, *border_gradient_b = NULL;
 	cairo_pattern_t *pattern = NULL;
 	GtkStateFlags state;
-	gint line_width;
+	gint line_width, focus_pad;
 	gint border_radius;
+	gboolean use_dashes;
+	double dashes[2] = { 2.0, 0.2 };
 
 	state = gtk_theming_engine_get_state (engine);
 	gtk_theming_engine_get (engine, state,
 				"-adwaita-focus-border-color", &border_color,
 				"-adwaita-focus-fill-color", &fill_color,
 				"-adwaita-focus-border-radius", &border_radius,
-				"-adwaita-focus-border-gradient-a", &border_gradient_a,
-				"-adwaita-focus-border-gradient-b", &border_gradient_b,
+				"-adwaita-focus-border-gradient", &pattern,
+				"-adwaita-focus-border-dashes", &use_dashes,
 				NULL);
 
 	gtk_theming_engine_get_style (engine,
 				      "focus-line-width", &line_width,
+				      "focus-padding", &focus_pad,
 				      NULL);
 
+	if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_NOTEBOOK) &&
+	    gtk_theming_engine_has_region (engine, GTK_STYLE_REGION_TAB, NULL)) {
+		/* as we render the tab smaller than the whole allocation, we need
+		 * to recenter and resize the focus on the tab.
+		 */
+		y += 3.0;
+		height -= 3.0;
+	}
+
 	cairo_save (cr);
 	cairo_set_line_width (cr, line_width);
 
@@ -173,44 +227,26 @@ adwaita_engine_render_focus (GtkThemingEngine *engine,
 		cairo_fill_preserve (cr);
 	}
 
+	if (use_dashes) {
+		cairo_set_dash (cr, dashes, 1, 0.0);
+	}
+
 	/* if we have a gradient, draw the gradient, otherwise
 	 * draw the line if we have a color for it.
 	 */
-	if (border_gradient_a != NULL &&
-	    border_gradient_b != NULL) {
-		pattern = cairo_pattern_create_linear (x, y, x, y + height);
-
-		cairo_pattern_add_color_stop_rgba (pattern,
-						   0.0,
-						   border_gradient_a->red,
-						   border_gradient_a->green,
-						   border_gradient_a->blue,
-						   border_gradient_a->alpha);
-
-		cairo_pattern_add_color_stop_rgba (pattern,
-						   1.0,
-						   border_gradient_b->red,
-						   border_gradient_b->green,
-						   border_gradient_b->blue,
-						   border_gradient_b->alpha);
+	if (pattern != NULL) {
+		style_pattern_set_matrix (pattern, width, height);
 
 		cairo_set_source (cr, pattern);
-		cairo_stroke (cr);
-
-		cairo_pattern_destroy (pattern);
 	} else if (border_color != NULL) {
 		gdk_cairo_set_source_rgba (cr, border_color);
-		cairo_stroke (cr);
 	}
 
+	cairo_stroke (cr);
 	cairo_restore (cr);
 
-	if (border_gradient_a != NULL) {
-		gdk_rgba_free (border_gradient_a);
-	}
-
-	if (border_gradient_b != NULL) {
-		gdk_rgba_free (border_gradient_b);
+	if (pattern != NULL) {
+		cairo_pattern_destroy (pattern);
 	}
 
 	if (border_color != NULL) {
@@ -329,7 +365,7 @@ adwaita_engine_render_check (GtkThemingEngine *engine,
 	}
 
 	if (render_from_assets_common (engine, cr,
-				       x, y, width, height)) {
+				       x, y + 2.0, width, height)) {
 		return;
 	}
 
@@ -451,7 +487,7 @@ adwaita_engine_render_option (GtkThemingEngine *engine,
 	}
 
 	if (render_from_assets_common (engine, cr,
-				       x, y, width, height)) {
+				       x, y + 2.0, width, height)) {
 		return;
 	}
 
@@ -575,7 +611,7 @@ render_notebook_extension (GtkThemingEngine *engine,
 			   GtkPositionType   gap_side)
 {
 	gint tab_curvature;
-	GdkRGBA *color, *notebook_border_color, border_color, background_color;
+	GdkRGBA *color, border_color, background_color;
 	GtkStateFlags state;
 	gdouble angle = 0;
 	cairo_pattern_t *pattern = NULL, *background_pattern = NULL;
@@ -589,7 +625,7 @@ render_notebook_extension (GtkThemingEngine *engine,
 	gtk_theming_engine_get_border_color (engine, state, &border_color);
 	gtk_theming_engine_get (engine, state,
 				"-adwaita-selected-tab-color", &color,
-				"-adwaita-notebook-border-color", &notebook_border_color,
+				"-adwaita-border-gradient", &pattern,
 				"background-image", &background_pattern,
 				NULL);
 
@@ -601,7 +637,18 @@ render_notebook_extension (GtkThemingEngine *engine,
 		cairo_translate (cr, width, height);
 	}
 
-	cairo_translate (cr, x + 0.5, y);
+	if (gap_side == GTK_POS_BOTTOM) {
+		cairo_translate (cr,
+				 x + 0.5,
+				 (state & GTK_STATE_FLAG_ACTIVE) ?
+				 y + 1.0 : y);
+	} else if (gap_side == GTK_POS_TOP) {
+		cairo_translate (cr,
+				 x - 0.5,
+				 (state & GTK_STATE_FLAG_ACTIVE) ?
+				 y - 1.0 : y);
+	}
+
 	cairo_rotate (cr, angle);
 
 	width -= 1.0;
@@ -628,20 +675,7 @@ render_notebook_extension (GtkThemingEngine *engine,
 	draw_tab_shape_active (cr, tab_curvature, 0, 0, width, height);
 
 	if (state & GTK_STATE_FLAG_ACTIVE) {
-		pattern = cairo_pattern_create_linear (0, height, 0, 0);
-		cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
-		cairo_pattern_add_color_stop_rgba (pattern, 0.0,
-						   notebook_border_color->red,
-						   notebook_border_color->green,
-						   notebook_border_color->blue,
-						   notebook_border_color->alpha);
-
-		cairo_pattern_add_color_stop_rgba (pattern, 1.0,
-						   border_color.red,
-						   border_color.green,
-						   border_color.blue,
-						   border_color.alpha);
-
+		style_pattern_set_matrix (pattern, width, height - 6.0);
 		cairo_set_source (cr, pattern);
 	} else {
 		gdk_cairo_set_source_rgba (cr, &border_color);
@@ -650,7 +684,6 @@ render_notebook_extension (GtkThemingEngine *engine,
 	cairo_stroke (cr);
 
 	gdk_rgba_free (color);
-	gdk_rgba_free (notebook_border_color);
 
 	if (pattern != NULL) {
 		cairo_pattern_destroy (pattern);
@@ -789,6 +822,63 @@ render_menubar_active_frame (GtkThemingEngine *engine,
 	gtk_border_free (border);
 }
 
+static void
+render_frame_default (GtkThemingEngine *engine,
+		      cairo_t *cr,
+		      gdouble x,
+		      gdouble y,
+		      gdouble width,
+		      gdouble height)
+{
+	cairo_pattern_t *pattern = NULL;
+	GtkStateFlags state;
+	GtkBorder *border;
+	gint line_width, border_radius;
+	GtkBorderStyle border_style;
+
+	state = gtk_theming_engine_get_state (engine);
+
+	gtk_theming_engine_get (engine, state,
+				"-adwaita-border-gradient", &pattern,
+				"border-style", &border_style,
+				NULL);
+
+	if (pattern == NULL || border_style == GTK_BORDER_STYLE_NONE) {
+		GTK_THEMING_ENGINE_CLASS (adwaita_engine_parent_class)->render_frame
+			(engine, cr,
+			 x, y, width, height);
+
+		return;
+	}
+
+	cairo_save (cr);
+
+	gtk_theming_engine_get (engine, state,
+				"border-radius", &border_radius,
+				"border-width", &border,
+				NULL);
+
+	line_width = MIN (MIN (border->top, border->bottom),
+			  MIN (border->left, border->right));
+	style_pattern_set_matrix (pattern, width, height);
+
+	cairo_set_line_width (cr, line_width);
+	_cairo_round_rectangle (cr,
+				x + line_width / 2.0,
+				y + line_width / 2.0,
+				width - line_width,
+				height - line_width,
+				border_radius);
+	cairo_set_source (cr, pattern);
+
+	cairo_stroke (cr);
+
+	cairo_restore (cr);
+
+	cairo_pattern_destroy (pattern);
+	gtk_border_free (border);
+}
+
 static void
 adwaita_engine_render_frame (GtkThemingEngine *engine,
 			     cairo_t          *cr,
@@ -884,8 +974,7 @@ adwaita_engine_render_frame (GtkThemingEngine *engine,
 			gtk_border_free (border_width);
 		}
 
-		GTK_THEMING_ENGINE_CLASS (adwaita_engine_parent_class)->render_frame (engine, cr, x, y,
-										      width, height);
+		render_frame_default (engine, cr, x, y, width, height);
 	}
 
 out:
@@ -1190,6 +1279,52 @@ draw_mark_slider (cairo_t *cr,
 	}
 }
 
+static void
+render_switch_lines (GtkThemingEngine *engine,
+		     cairo_t *cr,
+		     gdouble x,
+		     gdouble y,
+		     gdouble width,
+		     gdouble height,
+		     GtkOrientation orientation)
+{
+	GtkStateFlags state;
+	GdkRGBA *lines_color;
+
+	state = gtk_theming_engine_get_state (engine);
+
+	if (state & GTK_STATE_FLAG_INSENSITIVE) {
+		return;
+	}
+
+	gtk_theming_engine_get (engine, state,
+				"-adwaita-switch-grip-color", &lines_color,
+				NULL);
+
+	cairo_save (cr);
+
+	cairo_translate (cr,
+			 x + width / 2.0 - 4.0,
+			 y + height / 2.0 - 3.0);
+
+	cairo_move_to (cr, 0.0, 0.0);
+	cairo_set_line_width (cr, 2.0);
+	cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);
+
+	cairo_line_to (cr, 0.0, 6.0);
+	cairo_move_to (cr, 4.0, 0.0);
+	cairo_line_to (cr, 4.0, 6.0);
+	cairo_move_to (cr, 8.0, 0.0);
+	cairo_line_to (cr, 8.0, 6.0);
+
+	gdk_cairo_set_source_rgba (cr, lines_color);
+	cairo_stroke (cr);
+
+	cairo_restore (cr);
+
+	gdk_rgba_free (lines_color);
+}
+
 static void
 adwaita_engine_render_slider (GtkThemingEngine *engine,
 			      cairo_t          *cr,
@@ -1210,7 +1345,6 @@ adwaita_engine_render_slider (GtkThemingEngine *engine,
 		cairo_matrix_t matrix;
 		GtkStateFlags state;
 		GdkRGBA color;
-		gdouble shade_factor = 2.0 / 3.0;
 		gboolean marks_above = FALSE, marks_below = FALSE;
 
 		if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_SCALE_HAS_MARKS_ABOVE)) {
@@ -1234,48 +1368,47 @@ adwaita_engine_render_slider (GtkThemingEngine *engine,
 					"background-image", &pattern,
 					NULL);
 
-		cairo_matrix_init_scale (&matrix, 1 / width, 1 / height);
-		cairo_pattern_set_matrix (pattern, &matrix);
-		cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
-
-		cairo_set_source (cr, pattern);
-		cairo_fill_preserve (cr);
-
-		gtk_theming_engine_get_border_color (engine, state, &color);
-
-		if (orientation == GTK_ORIENTATION_HORIZONTAL) {
-			border_pattern = cairo_pattern_create_linear (0, 0, 0, height);
+		if (pattern != NULL) {
+			style_pattern_set_matrix (pattern, width, height);
+			cairo_set_source (cr, pattern);
 		} else {
-			border_pattern = cairo_pattern_create_linear (width, 0, 0, 0);
+			gtk_theming_engine_get_background_color (engine, state, &color);
+			gdk_cairo_set_source_rgba (cr, &color);
 		}
 
-		cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
+		cairo_fill_preserve (cr);
 
-		cairo_pattern_add_color_stop_rgba (border_pattern,
-						   marks_above ? 0.0 : 1.0,
-						   color.red / shade_factor,
-						   color.green / shade_factor,
-						   color.blue / shade_factor,
-						   color.alpha);
+		gtk_theming_engine_get (engine, state,
+					"-adwaita-border-gradient", &border_pattern,
+					NULL);
 
-		cairo_pattern_add_color_stop_rgba (border_pattern,
-						   marks_above ? 1.0 : 0.0,
-						   color.red,
-						   color.green,
-						   color.blue,
-						   color.alpha);
+		if (border_pattern != NULL) {
+			style_pattern_set_matrix (border_pattern, width, height);
+			cairo_set_source (cr, border_pattern);
+		} else {
+			gtk_theming_engine_get_border_color (engine, state, &color);
+			gdk_cairo_set_source_rgba (cr, &color);
+		}
 
-		cairo_set_source (cr, border_pattern);
 		cairo_stroke (cr);
 
-		cairo_pattern_destroy (pattern);
-		cairo_pattern_destroy (border_pattern);
+		if (pattern != NULL) {
+			cairo_pattern_destroy (pattern);
+		}
+
+		if (border_pattern != NULL) {
+			cairo_pattern_destroy (border_pattern);
+		}
 	}
 	else
 	{
 		GTK_THEMING_ENGINE_CLASS (adwaita_engine_parent_class)->render_slider (engine, cr,
 										       x, y, width, height,
 										       orientation);
+
+		if (gtk_widget_path_is_type (path, GTK_TYPE_SWITCH)) {
+			render_switch_lines (engine, cr, x, y, width, height, orientation);
+		}
 	}
 
 	cairo_restore (cr);
@@ -1357,9 +1490,45 @@ adwaita_engine_render_handle (GtkThemingEngine *engine,
 		}
 
 		cairo_restore (cr);
-	}
-	else
-	{
+	} else if (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_PANE_SEPARATOR)) {
+		GdkRGBA bg, fg;
+		GtkStateFlags state;
+		gdouble xx, yy;
+
+		state = gtk_theming_engine_get_state (engine);
+		gtk_theming_engine_get_background_color (engine, state, &bg);
+		gtk_theming_engine_get_color (engine, state, &fg);
+
+		cairo_save (cr);
+
+		cairo_rectangle (cr, x, y,
+				 width, height);
+		gdk_cairo_set_source_rgba (cr, &bg);
+		cairo_fill (cr);
+
+		gdk_cairo_set_source_rgba (cr, &fg);
+		cairo_set_line_width (cr, 2.0);
+		cairo_line_to (cr, x + width, y + height);
+		cairo_stroke (cr);
+
+		if (width > height) {
+			for (xx = x + width / 2 - 12; xx <= x + width / 2 + 12; xx += 6) {
+				cairo_arc (cr, xx, y + height / 2 - 1,
+					   1.5,
+					   0, G_PI * 2.0);
+				cairo_fill (cr);
+			}
+		} else {
+			for (yy = y + height / 2 - 12; yy <= y + height / 2 + 12; yy += 6) {
+				cairo_arc (cr, x + width / 2.0, yy,
+					   1.0,
+					   0, G_PI * 2.0);
+				cairo_fill (cr);
+			}
+		}
+
+		cairo_restore (cr);
+	} else {
 		GTK_THEMING_ENGINE_CLASS (adwaita_engine_parent_class)->render_handle (engine, cr,
 										       x, y, width, height);
 	}
@@ -1518,6 +1687,46 @@ adwaita_engine_render_icon_pixbuf (GtkThemingEngine    *engine,
 	return stated;
 }
 
+static void
+adwaita_engine_render_line (GtkThemingEngine *engine,
+			    cairo_t *cr,
+			    gdouble x0,
+			    gdouble y0,
+			    gdouble x1,
+			    gdouble y1)
+{
+	const GtkWidgetPath *path;
+
+	path = gtk_theming_engine_get_path (engine);
+
+	if ((gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_MARK) &&
+	     gtk_widget_path_is_type (path, GTK_TYPE_SCALE)) ||
+	    (gtk_theming_engine_has_class (engine, GTK_STYLE_CLASS_SEPARATOR) &&
+	     gtk_widget_path_is_type (path, GTK_TYPE_TREE_VIEW))) {
+		GtkStateFlags state;
+		GdkRGBA bg;
+
+		state = gtk_theming_engine_get_state (engine);
+		gtk_theming_engine_get_background_color (engine, state, &bg);
+
+		cairo_save (cr);
+
+		cairo_set_line_cap (cr, CAIRO_LINE_CAP_SQUARE);
+		cairo_set_line_width (cr, 1);
+
+		cairo_move_to (cr, x0 + 0.5, y0 + 0.5);
+		cairo_line_to (cr, x1 + 0.5, y1 + 0.5);
+
+		gdk_cairo_set_source_rgba (cr, &bg);
+		cairo_stroke (cr);
+
+		cairo_restore (cr);
+	} else {
+		GTK_THEMING_ENGINE_CLASS (adwaita_engine_parent_class)->render_line (engine, cr,
+										     x0, y0, x1, y1);
+	}
+}
+
 static void
 adwaita_engine_class_init (AdwaitaEngineClass *klass)
 {
@@ -1535,6 +1744,7 @@ adwaita_engine_class_init (AdwaitaEngineClass *klass)
 	engine_class->render_slider = adwaita_engine_render_slider;
 	engine_class->render_handle = adwaita_engine_render_handle;
 	engine_class->render_icon_pixbuf = adwaita_engine_render_icon_pixbuf;
+	engine_class->render_line = adwaita_engine_render_line;
 
 	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
 					      g_param_spec_boxed ("focus-border-color",
@@ -1548,15 +1758,10 @@ adwaita_engine_class_init (AdwaitaEngineClass *klass)
 								0, G_MAXINT, 0,
 								0));
 	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
-					      g_param_spec_boxed ("focus-border-gradient-a",
-								  "Focus border gradient A",
-								  "Focus border gradient A",
-								  GDK_TYPE_RGBA, 0));
-	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
-					      g_param_spec_boxed ("focus-border-gradient-b",
-								  "Focus border gradient B",
-								  "Focus border gradient B",
-								  GDK_TYPE_RGBA, 0));
+					      g_param_spec_boxed ("focus-border-gradient",
+								  "Focus border gradient",
+								  "Focus border gradient",
+								  CAIRO_GOBJECT_TYPE_PATTERN, 0));
 	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
 					      g_param_spec_boxed ("focus-fill-color",
 								  "Focus fill color",
@@ -1568,9 +1773,24 @@ adwaita_engine_class_init (AdwaitaEngineClass *klass)
 								  "Selected tab color",
 								  GDK_TYPE_RGBA, 0));
 	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
-					      g_param_spec_boxed ("notebook-border-color",
-								  "Notebook border color",
-								  "Notebook border color",
+					      g_param_spec_boxed ("border-gradient",
+								  "Border gradient",
+								  "Border gradient",
+								  CAIRO_GOBJECT_TYPE_PATTERN, 0));
+	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
+					      g_param_spec_boolean ("focus-border-dashes",
+								    "Focus border uses dashes",
+								    "Focus border uses dashes",
+								    FALSE, 0));
+	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
+					      g_param_spec_boxed ("menuitem-arrow-color",
+								  "Menuitem arrow color",
+								  "Menuitem arrow color",
+								  GDK_TYPE_RGBA, 0));
+	gtk_theming_engine_register_property (ADWAITA_NAMESPACE, NULL,
+					      g_param_spec_boxed ("switch-grip-color",
+								  "Switch grip color",
+								  "Switch grip color",
 								  GDK_TYPE_RGBA, 0));
 }
 
